/*
 * Tapiocaui app
 * Copyright (C) 2005 INdT
 * @author Renato Araujo Oliveira Filho <renato.filho@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <libxml/entities.h>
#include <libxml/SAX.h>
#include <libxml/parserInternals.h>
#include <glib/gstring.h>
#include <glib/gstrfuncs.h>
#include <glib/gmessages.h>
#include <glib/gfileutils.h>

#include "tapiocaui_db.h"
#include "tapiocaui_utils.h"

#define TAPIOCAUI_DB_VERSION "0.2"

enum
{
    DB_CHANGED,
    DB_BUDDY_ADDED,
    DB_BUDDY_REMOVED,
    LAST_SIGNAL
};

typedef struct
{
    guint in_unknown_elt;
    TapiocaUIDBEntryType entry_type;
    xmlParserCtxtPtr xmlctx;
    GString *prop_value;
    TapiocaUIBuddyData *entry;
    TapiocaUIDB *db;

    enum {
        PARSER_STATE_START,
        PARSER_STATE_DB,
        PARSER_STATE_ENTRY,
        PARSER_STATE_ENTRY_PROPERTY,
        PARSER_STATE_END
    } state;
} TapiocaUIDBContext;

struct _TapiocaUIDBEntryPrivate
{
    TapiocaUIDBEntryType type;
};

struct _TapiocaUIDBPrivate
{
    gchar *file_name;
    GList *entry_list[TAPIOCAUI_DB_ENTRY_LAST];
};

static void         tapiocaui_db_class_init          (TapiocaUIDBClass *k);
static void         tapiocaui_db_init                (TapiocaUIDB *self);
static void         tapiocaui_db_finalize            (GObject *object);
static void         tapiocaui_db_parse_file          (TapiocaUIDB* self);
static const gchar *tapiocaui_db_entry_type          (TapiocaUIDBEntryType type);
static void         tapiocaui_db_entry_presence_free (gpointer data);

static guint tapiocaui_db_signals[LAST_SIGNAL] = { 0 };
static GObjectClass *parent_class = NULL;

G_DEFINE_TYPE (TapiocaUIDB, tapiocaui_db, G_TYPE_OBJECT)

static void
tapiocaui_db_class_init (TapiocaUIDBClass *k)
{
    GObjectClass *object_class = G_OBJECT_CLASS (k);
    parent_class = g_type_class_peek_parent (k);
    object_class->finalize = tapiocaui_db_finalize;

    tapiocaui_db_signals[DB_CHANGED] =
        g_signal_new ("db-changed",
                      TAPIOCAUI_TYPE_DB,
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (TapiocaUIDBClass, db_changed),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE,
                      0);
    tapiocaui_db_signals[DB_BUDDY_ADDED] =
        g_signal_new ("db-buddy-added",
                      TAPIOCAUI_TYPE_DB,
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (TapiocaUIDBClass, db_buddy_added),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__POINTER,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_POINTER);
    tapiocaui_db_signals[DB_BUDDY_REMOVED] =
        g_signal_new ("db-buddy-removed",
                      TAPIOCAUI_TYPE_DB,
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (TapiocaUIDBClass, db_buddy_removed),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__POINTER,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_POINTER);
}

static void
tapiocaui_db_init (TapiocaUIDB *self)
{
    int index;
    self->priv = g_new0 (TapiocaUIDBPrivate, 1);
    for (index = 0; index < TAPIOCAUI_DB_ENTRY_LAST; ++index)
        self->priv->entry_list[index] = NULL;
}

static void
tapiocaui_db_finalize (GObject *object)
{
    TapiocaUIDB *self = TAPIOCAUI_DB (object);
    g_free (self->priv->file_name);
    g_free (self->priv);

    G_OBJECT_CLASS (parent_class)->finalize (object);
}

TapiocaUIDB* tapiocaui_db_new (const gchar* file)
{
    TapiocaUIDB *self;
    self = g_object_new (TAPIOCAUI_TYPE_DB, NULL);
    self->priv->file_name = g_build_filename (g_get_home_dir (),
                                              ".tapioca",
                                              file, NULL);
    return self;
}

gboolean
tapiocaui_db_load (TapiocaUIDB *self)
{
    tapiocaui_db_parse_file (self);
    return TRUE;
}

void
tapiocaui_db_commit (TapiocaUIDB *self)
{
    GString *save_path = NULL;
    FILE *f;
    GList *lst_entry;
    gint index;

    save_path = g_string_new (self->priv->file_name);

    g_string_append (save_path, ".tmp");
    f = fopen (save_path->str, "w");
    if (!f) {
        g_critical ("Can't save XML: %s", g_strerror (errno));
        goto out;
    }
    fprintf (f, "<?xml version=\"1.0\" standalone=\"yes\"?>\n");
    fprintf (f, "<tapiocauidb version=\"%s\">\n", TAPIOCAUI_DB_VERSION);
    for (index=0; index < TAPIOCAUI_DB_ENTRY_LAST; index++) {
        lst_entry = self->priv->entry_list[index];

        while (lst_entry) {
            TapiocaUIBuddyData *buddy = (TapiocaUIBuddyData *) lst_entry->data;
            fprintf (f, "<entry type=\"%s\">\n", tapiocaui_db_entry_type (buddy->priv->type));
            fprintf (f, "\t<id>%s</id>\n", buddy->id);
            fprintf (f, "\t<name>%s</name>\n", buddy->name);
            fprintf (f, "\t<group>%s</group>\n", buddy->group);
            fprintf (f, "\t<account-type>%s</account-type>\n", buddy->account_type);
            fprintf (f, "\t<photo>%s</photo>\n", buddy->photo);
            fprintf (f, "</entry>\n");
            lst_entry = lst_entry->next;
        }
    }
    if (fclose (f) != 0) {
        g_critical ("Can't close XML: %s", g_strerror (errno));
        unlink (save_path->str);
        goto out;
    }
    if (rename (save_path->str, self->priv->file_name) != 0) {
        g_critical ("Can't close XML: %s", g_strerror (errno));
        unlink (save_path->str);
    }
out:
    if (save_path) 
        g_string_free (save_path, FALSE);
}

void 
tapiocaui_db_insert_buddy (TapiocaUIDB *self,
                           TapiocaUIBuddyData *buddy)
{
    TapiocaUIDBEntryType type = buddy->priv->type;
    gchar *alias;

    g_return_if_fail (buddy != NULL);

    alias = g_strdup (buddy->name);
    alias = g_strstrip (alias);

    if (strlen (alias) == 0) {
        g_free (buddy->name);
        buddy->name = g_strdup (buddy->id);
    }

    g_free (alias);
        
    self->priv->entry_list[type] = g_list_append (self->priv->entry_list[type], buddy);
    g_signal_emit (G_OBJECT (self), tapiocaui_db_signals [DB_BUDDY_ADDED], 0, (gpointer)buddy);
}

GList *
tapiocaui_db_get_buddy_list (TapiocaUIDB *self,
                             TapiocaUIDBEntryType type)
{
    return self->priv->entry_list[type];
}

static void
tapiocaui_db_set_entry_value (TapiocaUIDBContext *ctx,
                              const gchar *name,
                              const gchar *value)
{
    if (!g_ascii_strcasecmp (name, "id")) {
        ctx->entry->id = g_strstrip (g_strdup (value));
    } else if (!g_ascii_strcasecmp (name, "name")) {
        ctx->entry->name = g_strstrip (g_strdup (value));
    } else if (!g_ascii_strcasecmp (name, "group")) {
        ctx->entry->group = g_strstrip (g_strdup (value));
    } else if (!g_ascii_strcasecmp (name, "photo")) {
        ctx->entry->photo = g_strstrip (g_strdup (value));
    } else if (!g_ascii_strcasecmp (name, "account-type")) {
        ctx->entry->account_type = g_strstrip (g_strdup (value));
    }
}

static void 
tapiocaui_db_parser_start_element (TapiocaUIDBContext *ctx,
                                   const char *name,
                                   const char **attrs)
{
    switch (ctx->state)
    {
        case PARSER_STATE_START:
            if (!g_ascii_strcasecmp(name, "tapiocauidb"))
                ctx->state = PARSER_STATE_DB;
            else
                ctx->in_unknown_elt++;
            break;

        case PARSER_STATE_DB:
            if (!g_ascii_strcasecmp(name, "entry")) {
                ctx->state = PARSER_STATE_ENTRY;
                for (; *attrs; attrs += 2) {
                    if (!strcmp (*attrs, "type")) {
                        const char *type_name = *(attrs + 1);
                        if (!strcmp (type_name, "buddy"))
                            ctx->entry_type = TAPIOCAUI_DB_ENTRY_BUDDY;
                        else if (!strcmp (type_name, "account"))
                            ctx->entry_type = TAPIOCAUI_DB_ENTRY_ACCOUNT;
                        else
                            return;
                    }
                }
                ctx->entry = tapiocaui_db_entry_new (ctx->entry_type);
            }
            else
                ctx->in_unknown_elt++;
            break;

        case PARSER_STATE_ENTRY:
            ctx->state = PARSER_STATE_ENTRY_PROPERTY;
            break;

        case PARSER_STATE_ENTRY_PROPERTY:
        case PARSER_STATE_END:
            break;
    }
}

static void
tapiocaui_db_parser_end_element (TapiocaUIDBContext* ctx,
                                 const char *name)
{
    if (ctx->in_unknown_elt > 0) {
        ctx->in_unknown_elt--;
        return;
    }

    switch (ctx->state)
    {
        case PARSER_STATE_START:
            ctx->state = PARSER_STATE_END;
            break;
        case PARSER_STATE_DB:
            ctx->state = PARSER_STATE_START;
            break;
        case PARSER_STATE_ENTRY:
            tapiocaui_db_insert_buddy (ctx->db, ctx->entry);
            ctx->state = PARSER_STATE_DB;
            break;
        case PARSER_STATE_ENTRY_PROPERTY:
            tapiocaui_db_set_entry_value (ctx, name, ctx->prop_value->str);
            ctx->state = PARSER_STATE_ENTRY;
            g_string_truncate (ctx->prop_value, 0);
            break;
        case PARSER_STATE_END:
            break;
    }
}

static void
tapiocaui_parser_characters (TapiocaUIDBContext *ctx,
                             const char *data,
                             guint len)
{
    switch (ctx->state)
    {
        case PARSER_STATE_ENTRY_PROPERTY:
            g_string_append_len (ctx->prop_value, data, len);
            break;
        case PARSER_STATE_START:
        case PARSER_STATE_DB:
        case PARSER_STATE_ENTRY:
        case PARSER_STATE_END:
            break;
    }
}

void 
tapiocaui_db_parse_file (TapiocaUIDB* self)
{
    xmlParserCtxtPtr ctxt;
    xmlSAXHandlerPtr sax_handler;

    sax_handler = g_new0 (xmlSAXHandler, 1);

    TapiocaUIDBContext *ctx = g_new0 (TapiocaUIDBContext, 1);
    ctx->prop_value = g_string_new ("");
    ctx->db = self;

    // initializing parser
    sax_handler->startElement = (startElementSAXFunc) tapiocaui_db_parser_start_element;
    sax_handler->endElement = (endElementSAXFunc) tapiocaui_db_parser_end_element;
    sax_handler->characters = (charactersSAXFunc) tapiocaui_parser_characters;
    xmlSubstituteEntitiesDefault (1);

    if (g_file_test (self->priv->file_name, G_FILE_TEST_EXISTS)) {
        ctxt = xmlCreateFileParserCtxt (self->priv->file_name);
        ctx->xmlctx = ctxt;
        xmlFree (ctxt->sax);
        ctxt->userData = ctx;
        ctxt->sax = sax_handler;
        xmlParseDocument (ctxt);
        ctxt->sax = NULL;
        xmlFreeParserCtxt (ctxt);
    }

    g_free (sax_handler);
    g_free (ctx);
}

GList *
tapiocaui_db_filter (TapiocaUIDB *self,
                     TapiocaUIDBEntryType type,
                     TapiocaUIDBField propid,
                     const gchar *value)
{
    GList *filter = NULL;
    GList *all = tapiocaui_db_get_buddy_list (self, type);
    if (value == NULL)
        return all;

    while (all) {
        TapiocaUIBuddyData *all_item = (TapiocaUIBuddyData *) all->data;
        switch (propid) 
        {
            case TAPIOCAUI_DB_FIELD_ID:
                if (g_strrstr  (all_item->id, value) != NULL) {
                  filter = g_list_append (filter, all_item);
                }
                break;
            case TAPIOCAUI_DB_FIELD_NAME:
                if (g_strrstr (all_item->name, value) != NULL) {
                  filter = g_list_append (filter, all_item);
                }
                break;
            case TAPIOCAUI_DB_FIELD_GROUP:
                if (g_strrstr (all_item->group, value) != NULL) {
                  filter = g_list_append (filter, all_item);
                }
                break;
            case TAPIOCAUI_DB_FIELD_PHOTO:
                if (g_strrstr  (all_item->photo, value) != NULL) {
                  filter = g_list_append (filter, all_item);
                }
                break;
            case TAPIOCAUI_DB_FIELD_ACCOUNT_TYPE:
                if (g_strrstr  (all_item->account_type, value) != NULL) {
                  filter = g_list_append (filter, all_item);
                }
                break;
            default:
                break;
        }
        all = all->next;
    }
    return filter;
}

TapiocaUIBuddyData *
tapiocaui_db_find_buddy_by_id (TapiocaUIDB *self,
                               TapiocaUIDBEntryType type,
                               const gchar *id)
{
    GList *all = tapiocaui_db_get_buddy_list (self, type);
    if (id == NULL) 
        return NULL;

    while (all) {
        TapiocaUIBuddyData *all_item = (TapiocaUIBuddyData *) all->data;
        if (g_strcasecmp (all_item->id, id) == 0)
            return all_item;
        all = all->next;
    }

    return NULL;
}

void
tapiocaui_db_remove_buddy (TapiocaUIDB *self,
                           TapiocaUIBuddyData *buddy)
{
    self->priv->entry_list[buddy->priv->type] =
        g_list_remove (self->priv->entry_list [buddy->priv->type],
                       (gconstpointer ) buddy);
    g_signal_emit (G_OBJECT (self), tapiocaui_db_signals [DB_BUDDY_REMOVED], 0, (gpointer)buddy);
}

void
tapiocaui_db_sync_buddy (TapiocaUIDB *self,
                         TapiocaUIBuddyData *buddy)
{
    TapiocaUIBuddyData *local_buddy = NULL;

    g_return_if_fail (self != NULL);
    g_return_if_fail (buddy != NULL);

    local_buddy = tapiocaui_db_find_buddy_by_id (self,
                                                 TAPIOCAUI_DB_ENTRY_BUDDY,
                                                 buddy->id);

    if (local_buddy) 
        if (g_strcasecmp (local_buddy->group, buddy->group) == 0) 
            tapiocaui_db_remove_buddy (self, local_buddy);

    tapiocaui_db_insert_buddy (self, buddy);
}

void
tapiocaui_db_notify_change (TapiocaUIDB *self)
{
    g_signal_emit (G_OBJECT (self), tapiocaui_db_signals [DB_CHANGED], 0);
}

static const gchar *
tapiocaui_db_entry_type (TapiocaUIDBEntryType type)
{
    switch (type)
    {
        case TAPIOCAUI_DB_ENTRY_BUDDY:
            return ("buddy");
        case TAPIOCAUI_DB_ENTRY_ACCOUNT:
            return ("account");
        default:
            g_warning ("Invalid entry type");
            return NULL;
    }
}

TapiocaUIBuddyData *
tapiocaui_db_entry_new (TapiocaUIDBEntryType type)
{
    TapiocaUIBuddyData *tmp = g_new0 (TapiocaUIBuddyData, 1);
    tmp->priv = g_new0 (TapiocaUIDBEntryPrivate, 1);
    tmp->priv->type = type;
    tmp->id = NULL;
    tmp->name = NULL;
    tmp->photo = NULL;
    tmp->account_type = NULL;
    tmp->group = NULL;
    tmp->subscription_state = TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_LOCAL;
    tmp->jids = g_hash_table_new_full (g_str_hash,
                                       g_str_equal,
                                       g_free,
                                       tapiocaui_db_entry_presence_free);
    return tmp;
}

void
tapiocaui_db_entry_free (TapiocaUIBuddyData* entry)
{
    g_return_if_fail (entry != NULL);

    if (entry->id)
        g_free (entry->id);
    if (entry->name)
        g_free (entry->name);
    if (entry->group)
        g_free (entry->group);
    if (entry->photo)
        g_free (entry->photo);
    if (entry->account_type)
        g_free (entry->account_type);
    if (entry->jids) 
        g_hash_table_destroy (entry->jids);
    if (entry->priv) 
        g_free (entry->priv);

    g_free (entry);
}

static void
tapiocaui_db_entry_presence_free (gpointer data)
{
    TapiocaUIBuddyPresence *tmp = (TapiocaUIBuddyPresence *) data;
    if (tmp->status_desc)
        g_free (tmp->status_desc);
}

gboolean
tapiocaui_db_voice_cap_hash_filter (gpointer key,
                                    gpointer value,
                                    gpointer user_data)
{
    TapiocaUIBuddyPresence *tmp = (TapiocaUIBuddyPresence *) value;

    return ((tmp->voice_cap));
}
