/*
 * Tapiocaui app
 * Copyright (C) 2005 INdT
 * @author Renato Araujo Oliveira Filho <renato.filho@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <libxml/entities.h>
#include <libxml/SAX.h>
#include <libxml/parserInternals.h>
#include <glib/gstring.h>
#include <glib/gstrfuncs.h>
#include <glib/gmessages.h>
#include <glib/gfileutils.h>
#include <glib/gi18n.h>
#include <glib/gstdio.h>
#include <glib/ghash.h>
#include <gtk/gtkmain.h>
#include <gtk/gtk.h>

#include <tapioca/client/tapioca.h>

#include <time.h>

#include "tapiocaui_utils.h"
#include "tapiocaui_shell.h"
#include "tapiocaui_mainwin.h"

#include <gst/gst.h>
#include <farsight/farsight-session.h>
#include <farsight/farsight-stream.h>
#include <farsight/farsight-transport.h>

#define TAPIOCA_PROTOCOL "xmpp"

typedef struct
{
    TapiocaUIBuddyData *buddy;
    TapiocaUICommunication *comm_win;
    GMutex *mutex;
    GList *text_sessions;
    TapiocaTextSession *current_text_session;
    TapiocaAudioSession *audio_session;
    FarsightSession *session;
    FarsightStream *stream;
    GstElement *srcvolume;
    /* GstElement *sinkvolume; */
    guint candidate_id;
    gboolean active_candidate_set;
    gboolean remote_codecs_set;
} TapiocaUIShellActiveComm;

struct _TapiocaUIShellPrivate
{
    TapiocaConnectionManager *tcm;
    TapiocaConnection *tc;
    TapiocaPresence *presence;
    TapiocaContactList *contact_list;
    TapiocaUIDB *db_buddy;
    TapiocaUIDB *db_account;
    GtkWidget *win;
    GHashTable *active_comm_list;
    TapiocaUIShellActiveComm *active_comm;
    gboolean in_use;
};

static const gchar *
presence_status_desc[] = { 
    "Offline",
    "Available",
    "Away From Computer",
    "Extended Away",
    "Busy",
    NULL 
};

// shell functions
static void        tapiocaui_shell_class_init                        (TapiocaUIShellClass *k);
static void        tapiocaui_shell_init                              (TapiocaUIShell *self);
static void        tapiocaui_shell_finalize                          (GObject *object);
static void        tapiocaui_shell_complete_login                    (TapiocaUIShell *self,
                                                                      const gchar *user);
static TapiocaUIShellActiveComm *
                   tapiocaui_shell_get_communication                 (TapiocaUIShell *self,
                                                                      const gchar *user_id);
static void        tapiocaui_shell_call_buddy                        (TapiocaUIShell *self,
                                                                      TapiocaUIShellActiveComm *active_comm);
static void        tapiocaui_shell_open_call                         (TapiocaUIShell *self,
                                                                      TapiocaUIShellActiveComm *comm);
static void        tapiocaui_shell_close_call                        (TapiocaUIShell *self,
                                                                      TapiocaUIShellActiveComm *comm);
static void        tapiocaui_shell_set_buddy_presence_priority       (TapiocaUIBuddyPresence *buddy_presence);

// session callbacks
static void        tapiocaui_shell_state_changed_cb                  (TapiocaConnection *tc,  
                                                                      guint state, 
                                                                      gpointer user_data);
static void        tapiocaui_shell_incoming_session_cb               (TapiocaConnection *tc, 
                                                                      const gchar *from,
                                                                      guint type,
                                                                      TapiocaSession *session,
                                                                      gpointer user_data);
static void        tapiocaui_shell_connection_error_cb               (TapiocaConnection *tc,
                                                                      guint code,
                                                                      gpointer user_data);
static void        tapiocaui_shell_session_member_added_cb           (TapiocaSession *session,
                                                                      const gchar *uri,
                                                                      gpointer user_data);
static void        tapiocaui_shell_session_member_removed_cb         (TapiocaSession *session,
                                                                      const gchar *uri,
                                                                      gpointer user_data);
static void        tapiocaui_shell_session_add_member_er_cb          (TapiocaSession *session,
                                                                      guint code,
                                                                      gpointer user_data);
static void        tapiocaui_shell_session_state_changed_cb          (TapiocaSession *session,
                                                                      guint state,
                                                                      gpointer user_data);
static void        tapiocaui_shell_session_error_signal_cb           (TapiocaSession *session,
                                                                      guint code,
                                                                      gpointer user_data);
static void        tapiocaui_shell_text_session_incoming_cb          (TapiocaTextSession *session,
                                                                      const gchar *from,
                                                                      guint type,
                                                                      const gchar *subject,
                                                                      const gchar *body,
                                                                      guint timestamp,
                                                                      gpointer user_data);
static void        tapiocaui_shell_text_session_derror_cb            (TapiocaTextSession *session,
                                                                      const gchar *from,
                                                                      guint type,
                                                                      const gchar *subject,
                                                                      const gchar *body,
                                                                      guint timestamp,
                                                                      guint code,
                                                                      gpointer user_data);
static gboolean    tapiocaui_shell_call_buddy_find                   (gpointer key,
                                                                      gpointer value,
                                                                      gpointer data);

// presence callbacks
static void        tapiocaui_shell_presence_status_up_cb             (TapiocaPresence *presence,
                                                                      const gchar     *contact,
                                                                      guint            status,
                                                                      const gchar     *status_desc,
                                                                      gboolean         voice_cap,
                                                                      gpointer         user_data);
static void        tapiocaui_shell_presence_registered_cb            (TapiocaPresence *presence,
                                                                      gpointer         user_data);
static void        tapiocaui_shell_presence_unregistered_cb          (TapiocaPresence *presence,
                                                                      gpointer         user_data);

// Contact list callbacks
static void        tapiocaui_shell_contact_list_contact_authorized_cb     (TapiocaContactList *tcl,
                                                                           const gchar        *addr,
                                                                           gpointer            user_data);
static void        tapiocaui_shell_contact_list_contact_denied_cb         (TapiocaContactList *tcl,
                                                                           const gchar        *addr,
                                                                           gpointer            user_data);
static void        tapiocaui_shell_contact_list_auth_requested_cb         (TapiocaContactList  *tcl,
                                                                           const gchar         *addr,
                                                                           gpointer             user_data);
static void        tapiocaui_shell_contact_list_contacts_retrieved_cb     (TapiocaContactList  *tcl,
                                                                           const gchar         *addr,
                                                                           const gchar         *alias,
                                                                           const gchar         *group,
                                                                           guint                state,
                                                                           gpointer             user_data);
static void        tapiocaui_shell_contact_list_contact_info_retrieved_cb (TapiocaContactList  *tcl,
                                                                           const gchar         *addr,
                                                                           gpointer            *info,
                                                                           gpointer             user_data);

// mainwin callbacks
static void        tapiocaui_shell_mainwin_login_cb              (TapiocaUIMainWin *win,
                                                                  const gchar *user,
                                                                  const gchar *proxy_password,
                                                                  gpointer user_data);
static void        tapiocaui_shell_mainwin_login_cancelled_cb    (TapiocaUIMainWin *win,
                                                                  gpointer user_data);
static void        tapiocaui_shell_mainwin_logout_cb             (TapiocaUIMainWin *win,
                                                                  gpointer user_data);
static void        tapiocaui_shell_mainwin_call_cb               (TapiocaUIMainWin *win,
                                                                  const gchar *user,
                                                                  gpointer user_data);
static void        tapiocaui_shell_mainwin_open_chat_cb          (TapiocaUIMainWin *win,
                                                                  const gchar *user,
                                                                  gpointer user_data);
static gboolean    tapiocaui_shell_mainwin_delete_cb             (GtkWidget *widget,
                                                                  GdkEvent *event,
                                                                  gpointer user_data);
static void        tapiocaui_shell_buddy_added_cb                (TapiocaUIMainWin *win,
                                                                  const gchar      *user,
                                                                  gpointer user_data);
static void        tapiocaui_shell_buddy_removed_cb              (TapiocaUIMainWin *win,
                                                                  const gchar      *user,
                                                                  gpointer user_data);
static void        tapiocaui_shell_buddy_updated_cb              (TapiocaUIMainWin *win,
                                                                  const gchar      *old_user,
                                                                  const gchar      *new_user,
                                                                  gpointer user_data);
static void        tapiocaui_shell_account_status_updated_cb     (TapiocaUIMainWin *win,
                                                                  guint status,
                                                                  const gchar *status_desc,
                                                                  gpointer user_data);
                                                                  
// communication callbacks                                                                  
static void        tapiocaui_shell_communication_volume_change_cb(TapiocaUICommunication *comm,
                                                                  guint volume,
                                                                  gpointer user_data);
static void        tapiocaui_shell_communication_quit_cb         (TapiocaUICommunication *comm,
                                                                  gpointer user_data);
static void        tapiocaui_shell_communication_input_mute_cb   (TapiocaUICommunication *comm,
                                                                  gpointer user_data);
static void        tapiocaui_shell_communication_input_unmute_cb (TapiocaUICommunication *comm,
                                                                  gpointer user_data);
static void        tapiocaui_shell_communication_call_answer_cb  (TapiocaUICommunication *win,
                                                                  const gchar *user,
                                                                  gpointer user_data);
static void        tapiocaui_shell_communication_call_reject_cb  (TapiocaUICommunication *win,
                                                                  const gchar *user,
                                                                  gpointer user_data);
static void        tapiocaui_shell_communication_call_disconnect_cb (TapiocaUICommunication *comm,
                                                                     const gchar *user,
                                                                     gpointer user_data);
static void        tapiocaui_shell_communication_call_make_cb    (TapiocaUICommunication *win,
                                                                  const gchar *user,
                                                                  gpointer user_data);
static void        tapiocaui_shell_communication_send_message_cb (TapiocaUICommunication *comm,
                                                                  const gchar *msg,
                                                                  gpointer user_data);
static void        tapiocaui_shell_communication_call_hold_cb    (TapiocaUICommunication *comm,
                                                                  const gchar *user,
                                                                  gpointer user_data);
static void        tapiocaui_shell_communication_call_unhold_cb  (TapiocaUICommunication *comm,
                                                                  const gchar *user,
                                                                  gpointer user_data);                                                                  
/* NAT traversal stuff */
static void        setup_rtp_session                             (TapiocaUIShell *self,
                                                                  TapiocaUIShellActiveComm *comm);
static void        setup_rtp_stream                              (TapiocaUIShell *self,
                                                                  TapiocaUIShellActiveComm *comm);

static void        close_rtp_session                             (TapiocaUIShell *self,
                                                                  TapiocaUIShellActiveComm *comm);
static void        close_rtp_stream                              (TapiocaUIShell *self,
                                                                  TapiocaUIShellActiveComm *comm);

static void        remote_capabilities_updated_cb                (TapiocaAudioSession *session,
                                                                  const gchar *cap,
                                                                  gpointer user_data);
static void        remote_candidate_added_cb                     (TapiocaAudioSession *audio_session, 
                                                                  const gchar *name,
                                                                  const gchar *protocol,
                                                                  const gchar *ip,
                                                                  guint port,
                                                                  gdouble preference,
                                                                  const gchar *username,
                                                                  const gchar *password,
                                                                  const gchar *type,
                                                                  guint generation,
                                                                  gpointer user_data);
static void        stream_error                                  (FarsightStream *stream, 
                                                                  FarsightStreamError error, 
                                                                  const gchar *debug, 
                                                                  gpointer user_data);
static void        session_error                                 (FarsightSession *stream, 
                                                                  FarsightSessionError error, 
                                                                  const gchar *debug, 
                                                                  gpointer user_data);
static void        new_active_candidate_pair                     (FarsightStream *stream, 
                                                                  gchar *native_candidate, 
                                                                  gchar *remote_candidate, 
                                                                  gpointer user_data);
static void        codec_changed                                 (FarsightStream *stream, 
                                                                  gint codec_id, 
                                                                  gpointer user_data);
static void        new_native_candidate                          (FarsightStream *stream, 
                                                                  const gchar *candidate_id, 
                                                                  gpointer user_data);
static void        native_candidates_prepared                    (FarsightStream *stream, 
                                                                  gpointer user_data);
static void        state_changed                                 (FarsightStream *stream, 
                                                                  FarsightStreamState state, 
                                                                  FarsightStreamDirection dir, 
                                                                  gpointer user_data);
static GList      *create_codecs_from_sdp                        (const gchar *sdp);
static gchar      *create_sdp_from_codecs                        (const GList *codecs);

G_DEFINE_TYPE (TapiocaUIShell, tapiocaui_shell, G_TYPE_OBJECT)

static void
tapiocaui_shell_class_init (TapiocaUIShellClass *k)
{
    GObjectClass *object_class = G_OBJECT_CLASS (k);
    tapiocaui_shell_parent_class = g_type_class_peek_parent (k);
    object_class->finalize = tapiocaui_shell_finalize;
}

static void
tapiocaui_shell_init (TapiocaUIShell *self)
{
    char *dir_name;

    self->priv = g_new0 (TapiocaUIShellPrivate, 1);

    // create conf dir
    dir_name = g_build_filename (g_get_home_dir (),
                                 ".tapioca", NULL);
    if (!g_file_test (dir_name, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) 
        if (g_mkdir (dir_name, 0750) != 0) 
            g_warning ("Error creating dir $HOME/.tapioca");

    g_free (dir_name);
}

static void
tapiocaui_shell_finalize (GObject *object)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (object);
    g_free (self->priv);

    G_OBJECT_CLASS (tapiocaui_shell_parent_class)->finalize (object);
}

TapiocaUIShell * 
tapiocaui_shell_new (void)
{
    TapiocaUIShell *self;
    self = g_object_new (TAPIOCAUI_TYPE_SHELL, NULL);
    return self;
}

gboolean
tapiocaui_shell_initialize (TapiocaUIShell *self)
{
    self->priv->db_account = tapiocaui_db_new ("accounts.xml");
    tapiocaui_db_load (self->priv->db_account);
    self->priv->win = GTK_WIDGET (tapiocaui_main_win_new (self->priv->db_account));
    self->priv->active_comm_list =  g_hash_table_new_full (g_str_hash, 
                                                      g_str_equal, 
                                                      g_free, 
                                                      NULL);

    g_signal_connect (G_OBJECT (self->priv->win),
                      "delete-event",
                      G_CALLBACK (tapiocaui_shell_mainwin_delete_cb),
                      NULL);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "login",
                      G_CALLBACK (tapiocaui_shell_mainwin_login_cb),
                      self);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "login-cancelled",
                      G_CALLBACK (tapiocaui_shell_mainwin_login_cancelled_cb),
                      self);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "logout",
                      G_CALLBACK (tapiocaui_shell_mainwin_logout_cb),
                      self);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "call",
                      G_CALLBACK (tapiocaui_shell_mainwin_call_cb),
                      self);
    g_signal_connect (G_OBJECT (self->priv->win),
                      "open-chat",
                      G_CALLBACK (tapiocaui_shell_mainwin_open_chat_cb),
                      self);

    // initialize tapioca
    if (!(self->priv->tcm = tapioca_connection_manager_new ())) {
        tapiocaui_debug ("Error on initialize tapioca");
        return FALSE;
    }
        
    return TRUE;
}

void
tapiocaui_shell_terminate (TapiocaUIShell *self)
{
    if (self->priv->tcm != NULL) {
        g_object_unref (self->priv->tcm);
        self->priv->tcm = NULL;
    }

    if (self->priv->tc != NULL) {
        g_object_unref (self->priv->tc);
        self->priv->tc = NULL;
    }

    g_hash_table_destroy (self->priv->active_comm_list);
    g_free (self->priv);
    g_object_unref (self->priv->db_account);
}

//
// Shell functions
//

static TapiocaUIShellActiveComm *
tapiocaui_shell_get_communication (TapiocaUIShell *self,
                                   const gchar *jid)
{
    TapiocaUIShellActiveComm *comm;
    TapiocaUIBuddyData *buddy;
    TapiocaUIBuddyPresence *buddy_presence;
    gchar *bare_jid;

    tapiocaui_debug ("searching for active communication for user %s on hash table %p", 
                     jid, self->priv->active_comm_list);

    if (strchr (jid, '/'))
        bare_jid = g_strndup (jid, strchr (jid, '/') - jid);
    else
        bare_jid = g_strdup (jid);

    comm = g_hash_table_lookup (self->priv->active_comm_list, bare_jid); 

    if (comm) {
        tapiocaui_debug ("communication found");
        g_free (bare_jid);
        return comm;
    }
   
    comm = g_new0 (TapiocaUIShellActiveComm, 1);
    comm->mutex = g_mutex_new ();

    tapiocaui_debug ("creating communication");
    buddy = tapiocaui_db_find_buddy_by_id (self->priv->db_buddy,
                                           TAPIOCAUI_DB_ENTRY_BUDDY,
                                           bare_jid);

    if (buddy && (g_hash_table_size (buddy->jids) == 0)) {
        g_free (bare_jid);
        return NULL;
    }

    if (buddy == NULL) {
        tapiocaui_debug ("unknown buddy calling");
        buddy = tapiocaui_db_entry_new (TAPIOCAUI_DB_ENTRY_BUDDY);
        buddy->name = g_strdup (bare_jid);
        // buddy->jids = g_list_append (buddy->jids, jid);
        buddy->id = g_strdup (bare_jid);
        buddy->group = g_strdup ("");
        buddy_presence = g_new0 (TapiocaUIBuddyPresence, 1);
        buddy_presence->status = TAPIOCA_PRESENCE_STATUS_AVAILABLE;
        buddy_presence->status_desc = g_strdup (presence_status_desc [1]);
        tapiocaui_shell_set_buddy_presence_priority (buddy_presence);
        g_hash_table_insert (buddy->jids, g_strdup (jid), buddy_presence);
    }
    tapiocaui_debug ("request chat window");
    comm->comm_win = tapiocaui_main_win_open_chat (TAPIOCAUI_MAIN_WIN (self->priv->win), 
                                                   buddy);
    comm->buddy = buddy;
    tapiocaui_debug ("insert communication %p with user %s on hash table %p", 
                     comm, bare_jid, self->priv->active_comm_list);
    g_hash_table_insert (self->priv->active_comm_list, g_strdup (bare_jid), comm);

    g_signal_connect (G_OBJECT (comm->comm_win),
                      "send-message",
                      G_CALLBACK (tapiocaui_shell_communication_send_message_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-answer",
                      G_CALLBACK (tapiocaui_shell_communication_call_answer_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-reject",
                      G_CALLBACK (tapiocaui_shell_communication_call_reject_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-disconnect",
                      G_CALLBACK (tapiocaui_shell_communication_call_disconnect_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-make",
                      G_CALLBACK (tapiocaui_shell_communication_call_make_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "quit",
                      G_CALLBACK (tapiocaui_shell_communication_quit_cb),
                      self);
 
    g_free (bare_jid);
        
    return comm;
}

static void
tapiocaui_shell_complete_login (TapiocaUIShell *self,
                                const gchar *user)
{
    gchar *db_name;

    db_name = g_strconcat (user, "_db.xml", NULL);
    self->priv->db_buddy = tapiocaui_db_new (db_name);
    tapiocaui_db_load (self->priv->db_buddy);

    /* Resquest the remote contact list */
    tapioca_contact_list_request_contacts (self->priv->contact_list);

    tapiocaui_main_win_load_buddy (TAPIOCAUI_MAIN_WIN (self->priv->win), 
                                   self->priv->db_buddy,
                                   user);

    g_free (db_name);
}

static void 
tapiocaui_shell_call_buddy (TapiocaUIShell *self,
                            TapiocaUIShellActiveComm *active_comm)
{
    TapiocaAudioSession *session = NULL;
    gchar *local_cap = NULL;

    g_return_if_fail (self != NULL);
    g_return_if_fail (active_comm != NULL); 

    session = (TapiocaAudioSession *) tapioca_connection_create_session (self->priv->tc, TAPIOCA_SESSION_TYPE_AUDIO);
    
    tapiocaui_debug ("audio session created : %p", session);

    g_assert (session != NULL);

    /* Connect session signals */
    g_signal_connect (G_OBJECT (session),
                      "member-added",
                      G_CALLBACK (tapiocaui_shell_session_member_added_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "member-removed",
                      G_CALLBACK (tapiocaui_shell_session_member_removed_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "add-member-error",
                      G_CALLBACK (tapiocaui_shell_session_add_member_er_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "state-changed",
                      G_CALLBACK (tapiocaui_shell_session_state_changed_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "error",
                      G_CALLBACK (tapiocaui_shell_session_error_signal_cb),
                      self);

    active_comm->audio_session = session;
 
    setup_rtp_session (self, active_comm);
    setup_rtp_stream (self, active_comm);
    farsight_stream_prepare_transports (active_comm->stream);
 
    local_cap = create_sdp_from_codecs (farsight_stream_get_local_codecs (active_comm->stream));
    
    tapiocaui_debug ("set local capabilities %s", local_cap);

    if (!(tapioca_audio_session_set_local_capabilities (
                    active_comm->audio_session,
                    (const gchar *) local_cap))) {
        tapiocaui_debug ("error on set local capabilities");
        g_object_unref (active_comm->audio_session);
        active_comm->audio_session = NULL;
        g_object_unref (active_comm->session);
        active_comm->session = NULL;
        g_object_unref (active_comm->stream);
        active_comm->stream = NULL;
        g_free (local_cap);
        return;
    }
        
    g_free (local_cap);

    /* Connect audio session signals */
    g_signal_connect (G_OBJECT (active_comm->audio_session),
                      "remote-capabilities-updated",
                      G_CALLBACK (remote_capabilities_updated_cb),
                      active_comm);
    g_signal_connect (G_OBJECT (active_comm->audio_session),
                      "remote-candidate-added",
                      G_CALLBACK (remote_candidate_added_cb),
                      active_comm);
    
    if (!g_hash_table_find (active_comm->buddy->jids, 
                       tapiocaui_shell_call_buddy_find, 
                       (gpointer) active_comm->audio_session)) {
        tapiocaui_debug ("buddy cannot receive calls");
        g_object_unref (active_comm->audio_session);
        active_comm->audio_session = NULL;
        g_object_unref (active_comm->session);
        active_comm->session = NULL;
        g_object_unref (active_comm->stream);
        active_comm->stream = NULL;
        return;
    }
    
    tapiocaui_communication_set_status (active_comm->comm_win,
            TAPIOCAUI_CALL_RINGING);
}

static void 
tapiocaui_shell_open_call (TapiocaUIShell *self, 
                           TapiocaUIShellActiveComm *comm)
{
    tapiocaui_debug ("open call %p", comm->audio_session);

    self->priv->in_use = TRUE;
    
    if (self->priv->active_comm) {
        tapiocaui_debug ("hold active call");
        // tapioca_audio_session_hold (self->priv->comm->audio_session);
        tapiocaui_communication_set_status (self->priv->active_comm->comm_win, TAPIOCAUI_CALL_HOLD);
        self->priv->active_comm = NULL; 
    }

    self->priv->active_comm = comm;

    tapiocaui_communication_set_status (comm->comm_win,
                                        TAPIOCAUI_CALL_ANSWERED);
    tapiocaui_communication_set_volume (comm->comm_win, 50);

    g_signal_connect (G_OBJECT (comm->comm_win),
                      "volume-change",
                      G_CALLBACK (tapiocaui_shell_communication_volume_change_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "input-sound-mute",
                      G_CALLBACK (tapiocaui_shell_communication_input_mute_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "input-sound-unmute",
                      G_CALLBACK (tapiocaui_shell_communication_input_unmute_cb),
                      self);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-hold",
                      G_CALLBACK (tapiocaui_shell_communication_call_hold_cb),
                      comm->audio_session);
    g_signal_connect (G_OBJECT (comm->comm_win),
                      "call-unhold",
                      G_CALLBACK (tapiocaui_shell_communication_call_unhold_cb),
                      comm->audio_session);
}

static void
tapiocaui_shell_close_call (TapiocaUIShell *self,
                            TapiocaUIShellActiveComm *comm)
{
    tapiocaui_communication_set_status (comm->comm_win, TAPIOCAUI_CALL_IDLE);

    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_volume_change_cb),
                                          self);
    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_input_mute_cb),
                                          comm->audio_session);
    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_input_unmute_cb),
                                          comm->audio_session);
    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_call_hold_cb),
                                          comm->audio_session);
    g_signal_handlers_disconnect_by_func (G_OBJECT (comm->comm_win),
                                          G_CALLBACK (tapiocaui_shell_communication_call_unhold_cb),
                                          comm->audio_session);

    if (comm->stream) 
        close_rtp_stream (self, comm);
    
    if (comm->session) 
        close_rtp_session (self, comm);

            
    if (comm->audio_session) {
        /* Check if session is closed before trying to close */
        if (tapioca_session_get_state (TAPIOCA_SESSION (comm->audio_session)) != 
            TAPIOCA_SESSION_STATE_CLOSED) {
            tapiocaui_debug ("closing audio session ...");
            tapioca_session_close (TAPIOCA_SESSION (comm->audio_session));
        }
        else
            comm->audio_session = NULL;
    }
    
    comm->active_candidate_set = FALSE;
    comm->remote_codecs_set = FALSE;
    
    self->priv->active_comm = NULL;
    self->priv->in_use = FALSE;
}

//
// main window callbacks
//
static gboolean
tapiocaui_shell_mainwin_delete_cb (GtkWidget *widget,
                                   GdkEvent *event,
                                   gpointer user_data)
{
    gtk_main_quit ();
    return FALSE;
}

static void
tapiocaui_shell_mainwin_login_cb (TapiocaUIMainWin *win,
                                  const gchar *user,
                                  const gchar *proxy_password,
                                  gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    GHashTable *table = NULL;
 
    self->priv->tc = tapioca_connection_manager_create_connection (
            self->priv->tcm, TAPIOCA_PROTOCOL);
    if (self->priv->tc == NULL) { 
        tapiocaui_main_win_cancel_login (win, -1);
        return;
    }
 
    table = g_hash_table_new_full (g_str_hash,
                                   g_str_equal,
                                   NULL,
                                  g_free);

    g_hash_table_insert (table, "username", g_strdup (user));
    g_hash_table_insert (table, "password", g_strdup (proxy_password));
    g_hash_table_insert (table, "server-address", g_strdup ("talk.google.com"));
    g_hash_table_insert (table, "server-port",  g_strdup ("5222"));
    g_hash_table_insert (table, "use-encryption", g_strdup ("1"));

    g_signal_connect (G_OBJECT (self->priv->tc),
                      "state-changed",
                      G_CALLBACK (tapiocaui_shell_state_changed_cb),
                      self);
    g_signal_connect (G_OBJECT (self->priv->tc),
                      "error",
                      G_CALLBACK (tapiocaui_shell_connection_error_cb),
                      self);

    if (!(tapioca_connection_connect (self->priv->tc, user, table))) {
        tapiocaui_debug ("error on tapioca_connect");
        tapiocaui_main_win_cancel_login (win, -1);
        g_hash_table_destroy (table);
        g_object_unref (G_OBJECT (self->priv->tc));
        self->priv->tc = NULL;
        return;
    }        

    g_hash_table_destroy (table);
    
    /* Check if remote contact list is supported */
    if (!(self->priv->contact_list = TAPIOCA_CONTACT_LIST (tapioca_connection_get_contact_list (self->priv->tc)))) {
        tapiocaui_debug ("Connection does not support remote contact list");
    }
    else {
        /* Connect all window signal for remote contact list */
        // TODO - remote contact list signals

        /* Connect all contact list signals */
        g_signal_connect (G_OBJECT (self->priv->contact_list),
                          "contact-authorized",
                          G_CALLBACK (tapiocaui_shell_contact_list_contact_authorized_cb),
                          self);
        g_signal_connect (G_OBJECT (self->priv->contact_list),
                          "contact-denied",
                          G_CALLBACK (tapiocaui_shell_contact_list_contact_denied_cb),
                          self);
        g_signal_connect (G_OBJECT (self->priv->contact_list),
                          "authorization-requested",
                          G_CALLBACK (tapiocaui_shell_contact_list_auth_requested_cb),
                          self);
        g_signal_connect (G_OBJECT (self->priv->contact_list),
                          "contacts-retrieved",
                          G_CALLBACK (tapiocaui_shell_contact_list_contacts_retrieved_cb),
                          self);
        g_signal_connect (G_OBJECT (self->priv->contact_list),
                          "contact-info-retrieved",
                          G_CALLBACK (tapiocaui_shell_contact_list_contact_info_retrieved_cb),
                          self);
    }

    /* Check if presence is supported */
    if (!(self->priv->presence = TAPIOCA_PRESENCE (tapioca_connection_get_presence (self->priv->tc)))) {
        tapiocaui_debug ("Connection does not support presence");
    }
    else {
        /* Connect all window signals for presence */
        g_signal_connect (G_OBJECT (self->priv->win),
                          "buddy-added",
                          G_CALLBACK (tapiocaui_shell_buddy_added_cb),
                          self);
        g_signal_connect (G_OBJECT (self->priv->win),
                          "buddy-removed",
                          G_CALLBACK (tapiocaui_shell_buddy_removed_cb),
                          self);
        g_signal_connect (G_OBJECT (self->priv->win),
                          "buddy-updated",
                          G_CALLBACK (tapiocaui_shell_buddy_updated_cb),
                          self);
        g_signal_connect (G_OBJECT (self->priv->win),
                          "account-status-updated",
                          G_CALLBACK (tapiocaui_shell_account_status_updated_cb),
                          self);

        /* Connect all presence signals */
        g_signal_connect (G_OBJECT (self->priv->presence),
                          "status-updated",
                          G_CALLBACK (tapiocaui_shell_presence_status_up_cb),
                          self);
        g_signal_connect (G_OBJECT (self->priv->presence),
                          "registered",
                          G_CALLBACK (tapiocaui_shell_presence_registered_cb),
                          self);
        g_signal_connect (G_OBJECT (self->priv->presence),
                          "unregistered",
                          G_CALLBACK (tapiocaui_shell_presence_unregistered_cb),
                          self);
    }
}

static void
tapiocaui_shell_mainwin_login_cancelled_cb (TapiocaUIMainWin *win, 
                                            gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    g_object_unref (self->priv->tc);
    self->priv->tc = NULL;
}

static void
tapiocaui_shell_mainwin_logout_cb (TapiocaUIMainWin *win,
                                   gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    /* Disconnect all window signals for presence */
    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->win),
                                          G_CALLBACK (tapiocaui_shell_buddy_added_cb),
                                          self);
    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->win),
                                          G_CALLBACK (tapiocaui_shell_buddy_removed_cb),
                                          self);
    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->win),
                                          G_CALLBACK (tapiocaui_shell_buddy_updated_cb),
                                          self);
    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->win),
                                          G_CALLBACK (tapiocaui_shell_account_status_updated_cb),
                                          self);

    /* Disconnect all presence signals */
    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->presence),
                                          G_CALLBACK (tapiocaui_shell_presence_status_up_cb),
                                          self);
    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->presence),
                                          G_CALLBACK (tapiocaui_shell_presence_registered_cb),
                                          self);
    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->presence),
                                          G_CALLBACK (tapiocaui_shell_presence_unregistered_cb),
                                          self);

    /* Disconnect all tapioca connection signals */
    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->tc),
                                          G_CALLBACK (tapiocaui_shell_state_changed_cb),
                                          self);

    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->tc),
                                          G_CALLBACK (tapiocaui_shell_connection_error_cb),
                                          self);

    if (self->priv->tc) {
        g_object_unref (self->priv->tc);
        self->priv->tc = NULL;
    }

    g_object_unref (self->priv->db_buddy);
    tapiocaui_main_win_logout (TAPIOCAUI_MAIN_WIN (self->priv->win));
}

static void
tapiocaui_shell_mainwin_call_cb (TapiocaUIMainWin *win,
                                 const gchar *user,
                                 gpointer user_data)
{
    TapiocaUIShellActiveComm *active_comm;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    tapiocaui_debug ("trying to make a call");
    active_comm = tapiocaui_shell_get_communication (self, user);
    tapiocaui_shell_call_buddy (self, active_comm);    
}


static void
tapiocaui_shell_mainwin_open_chat_cb (TapiocaUIMainWin *win,
                                      const gchar *user,
                                      gpointer user_data)
{
    TapiocaUIShellActiveComm *comm;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    TapiocaUIBuddyData *buddy = NULL;

    tapiocaui_debug ("");

    buddy = tapiocaui_db_find_buddy_by_id (self->priv->db_buddy,
                                           TAPIOCAUI_DB_ENTRY_BUDDY,
                                           user);

    comm = tapiocaui_shell_get_communication (self, user);
}

//
//communication callbacks
//
static void 
tapiocaui_shell_communication_send_message_cb (TapiocaUICommunication *comm,
                                               const gchar *msg,
                                               gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data); 
    TapiocaUIShellActiveComm *active_comm;
    gchar **members = NULL;
    active_comm = g_hash_table_lookup (self->priv->active_comm_list, tapiocaui_communication_get_buddy (comm)->id); 
    const gchar *subject = "";
    TapiocaTextSession *text_session;
    
    if (!active_comm) {
        tapiocaui_debug ("error on send msg");
        return;
    }

    if (!(active_comm->text_sessions)) {
        tapiocaui_debug ("creating a new text session and add members"); 
        text_session = (TapiocaTextSession *) tapioca_connection_create_session (self->priv->tc, 
                TAPIOCA_SESSION_TYPE_TEXT);

        active_comm->text_sessions = g_list_append (active_comm->text_sessions, text_session);

        /* Connect session signals */
        g_signal_connect (G_OBJECT (text_session),
                          "member-added",
                          G_CALLBACK (tapiocaui_shell_session_member_added_cb),
                          self);
        g_signal_connect (G_OBJECT (text_session),
                          "member-removed",
                          G_CALLBACK (tapiocaui_shell_session_member_removed_cb),
                          self);
        g_signal_connect (G_OBJECT (text_session),
                          "add-member-error",
                          G_CALLBACK (tapiocaui_shell_session_add_member_er_cb),
                          self);
        g_signal_connect (G_OBJECT (text_session),
                          "state-changed",
                          G_CALLBACK (tapiocaui_shell_session_state_changed_cb),
                          self);
        g_signal_connect (G_OBJECT (text_session),
                          "error",
                          G_CALLBACK (tapiocaui_shell_session_error_signal_cb),
                          self);

        /* Connect text session signals */
        g_signal_connect (G_OBJECT (text_session),
                          "incoming",
                          G_CALLBACK (tapiocaui_shell_text_session_incoming_cb),
                          self);

        g_signal_connect (G_OBJECT (text_session),
                          "delivery-error",
                          G_CALLBACK (tapiocaui_shell_text_session_derror_cb),
                          self);
    
        tapioca_text_session_disable_message_queue (text_session);

        members = g_new0 (gchar *, 2);
        members[0] = tapiocaui_communication_get_buddy (comm)->id;
        tapioca_session_add_members (TAPIOCA_SESSION (text_session), (const gchar **) members);

        if (members) {
            g_free (members);
            members = NULL;
        }

        tapioca_session_initiate (TAPIOCA_SESSION (text_session));

        active_comm->current_text_session = text_session;
    }

    tapiocaui_debug ("sending message [%s] [%s]", tapiocaui_communication_get_buddy (comm)->id, msg); 

    tapioca_text_session_send (active_comm->current_text_session,
            TAPIOCA_TEXT_SESSION_MESSAGE_TYPE_PLAIN_TEXT,
            subject,
            msg);
}

//accept incomming call
static void
tapiocaui_shell_communication_call_answer_cb (TapiocaUICommunication *comm,
                                              const gchar *user,
                                              gpointer user_data)
{
    TapiocaUIShellActiveComm *active_comm;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    gchar *local_cap = NULL;

    active_comm = tapiocaui_shell_get_communication (self, user);

    local_cap = create_sdp_from_codecs (farsight_stream_get_local_codecs (active_comm->stream));
    
    tapiocaui_debug ("set local capabilities %s", local_cap);

    if (!(tapioca_audio_session_set_local_capabilities (
                    active_comm->audio_session,
                    (const gchar *) local_cap))) {
        tapiocaui_debug ("error on set local cap");
        g_object_unref (active_comm->audio_session);
        active_comm->audio_session = NULL;
        g_free (local_cap);
        return;
    }
        
    g_free (local_cap);

    if (!tapioca_audio_session_accept (active_comm->audio_session)) {
        tapiocaui_communication_set_status (active_comm->comm_win, 
                                            TAPIOCAUI_CALL_IDLE);
    }
}

//reject incomming call
static void
tapiocaui_shell_communication_call_reject_cb (TapiocaUICommunication *comm,
                                              const gchar *user,
                                              gpointer user_data)
{
    TapiocaUIShellActiveComm *active_comm;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    active_comm = tapiocaui_shell_get_communication (self, user);

    tapiocaui_debug ("Rejecting a call ...");

    if (!(tapioca_audio_session_reject (active_comm->audio_session)))
        tapiocaui_debug ("Failed on reject a call");

    tapiocaui_shell_close_call (self, active_comm);
}

//disconnect incomming call
static void
tapiocaui_shell_communication_call_disconnect_cb (TapiocaUICommunication *comm,
                                                  const gchar *user,
                                                  gpointer user_data)
{
    TapiocaUIShellActiveComm *active_comm;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    active_comm = tapiocaui_shell_get_communication (self, user);
    tapiocaui_shell_close_call (self, active_comm);
}

//make a call
static void
tapiocaui_shell_communication_call_make_cb (TapiocaUICommunication *comm,
                                            const gchar *user,
                                            gpointer user_data)
{
    TapiocaUIShellActiveComm *active_comm;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    tapiocaui_debug ("trying to make a call");
    active_comm = tapiocaui_shell_get_communication (self, user);
    tapiocaui_shell_call_buddy (self, active_comm);    
}

static void
tapiocaui_shell_communication_call_hold_cb (TapiocaUICommunication *comm,
                                            const gchar *user,
                                            gpointer user_data)
{
    tapiocaui_debug ("hold not implemented");
    // tapiocaui_communication_set_status (comm, TAPIOCAUI_CALL_HOLD);    
}

static void
tapiocaui_shell_communication_call_unhold_cb (TapiocaUICommunication *comm,
                                              const gchar *user,
                                              gpointer user_data)
{
    tapiocaui_debug ("unhold not implemented");
    // tapiocaui_communication_set_status (comm, TAPIOCAUI_CALL_ANSWERED);
}

// when communication window is closed
static void
tapiocaui_shell_communication_quit_cb (TapiocaUICommunication *comm,
                                       gpointer user_data)
{
    TapiocaUIShellActiveComm *active_comm;
    const TapiocaUIBuddyData *buddy;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    GList *text_sessions;
    
    tapiocaui_debug ("communication window closed");
    buddy = tapiocaui_communication_get_buddy (comm);
    active_comm = tapiocaui_shell_get_communication (self, buddy->id);
   
    tapiocaui_shell_close_call (self, active_comm);
    
    for (text_sessions = active_comm->text_sessions; text_sessions; text_sessions = g_list_next (text_sessions)) {
        g_object_unref ((TapiocaTextSession *) text_sessions->data);
    }
    g_list_free (active_comm->text_sessions);
    active_comm->text_sessions = NULL;

    g_hash_table_remove (self->priv->active_comm_list, buddy->id);
    
    g_mutex_free (active_comm->mutex);

    g_free (active_comm);
    self->priv->active_comm = NULL;
}

static void
tapiocaui_shell_communication_volume_change_cb (TapiocaUICommunication *comm,
                                                guint volume,
                                                gpointer user_data)
{
    /* 
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    if (volume > 100)
        volume = 100;
    tapiocaui_debug ("volume change to %d", volume);
    g_object_set (G_OBJECT (self->priv->active_comm->sinkvolume), "volume", (double) volume / 25, NULL); 
    */
    
    tapiocaui_debug ("Not implemented");
}

static void
tapiocaui_shell_communication_input_mute_cb (TapiocaUICommunication *comm,
                                             gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    tapiocaui_debug ("input mute");
    g_object_set (G_OBJECT (self->priv->active_comm->srcvolume), "mute", TRUE, NULL);
}

static void
tapiocaui_shell_communication_input_unmute_cb (TapiocaUICommunication *comm,
                                               gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    tapiocaui_debug ("input unmute");
    g_object_set (G_OBJECT (self->priv->active_comm->srcvolume), "mute", FALSE, NULL);
}

/* here will be make all state change process */
static void
tapiocaui_shell_state_changed_cb (TapiocaConnection *tc,
                                  guint state,
                                  gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    guint old_state;

    switch (state) {
        case TAPIOCA_CONNECTION_STATE_DISCONNECTED: 
            tapiocaui_debug ("STATE_CONNECTION_DISCONNECTED");
            old_state = tapioca_connection_get_state (tc);
            if (old_state == TAPIOCA_CONNECTION_STATE_CONNECTING)
                tapiocaui_main_win_cancel_login (TAPIOCAUI_MAIN_WIN(self->priv->win), -1);
            else
                tapiocaui_main_win_logout (TAPIOCAUI_MAIN_WIN (self->priv->win));
            break;
        
        case TAPIOCA_CONNECTION_STATE_CONNECTED: {
            tapiocaui_debug ("STATE_CONNECTION_CONNECTED");
            g_signal_connect (G_OBJECT (self->priv->tc),
                              "incoming-session",
                              G_CALLBACK (tapiocaui_shell_incoming_session_cb),
                              self);

            tapiocaui_shell_complete_login (self, tapioca_connection_get_id (tc));
           
            if (self->priv->presence) {
                guint i = 0;
                gchar **contacts = NULL;
                const gchar *status_desc = NULL;
                GList *buddy_list;
                
                tapioca_presence_subscribe (self->priv->presence, TAPIOCA_PRESENCE_STATUS_AVAILABLE, status_desc);

                buddy_list = tapiocaui_db_get_buddy_list (self->priv->db_buddy, 
                                                          TAPIOCAUI_DB_ENTRY_BUDDY);
                if (buddy_list == NULL)
                    return;

                contacts = g_new0 (gchar *, g_list_length (buddy_list) + 1);
                for (; buddy_list; buddy_list = g_list_next (buddy_list)) 
                    contacts[i++] = g_strdup (((TapiocaUIBuddyData *) buddy_list->data)->id);
                tapioca_presence_request_status (self->priv->presence, (const gchar **) contacts);
                g_strfreev (contacts);
            }

            break;
        }

        case TAPIOCA_CONNECTION_STATE_CONNECTING:
            tapiocaui_debug ("STATE_CONNECTION_CONNECTING");
        default:
            break;
    }
}

/* here will be processed all incoming session */
static void
tapiocaui_shell_incoming_session_cb (TapiocaConnection *tc,
                                     const gchar *from,
                                     guint type,
                                     TapiocaSession *session,
                                     gpointer user_data)
{
    TapiocaUIShellActiveComm *comm;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    if (type == TAPIOCA_SESSION_TYPE_AUDIO) {
        if (self->priv->in_use) {
            tapioca_audio_session_reject (TAPIOCA_AUDIO_SESSION (session));
            return;
        }
    }

    /* Connect session signals */
    g_signal_connect (G_OBJECT (session),
                      "member-added",
                      G_CALLBACK (tapiocaui_shell_session_member_added_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "member-removed",
                      G_CALLBACK (tapiocaui_shell_session_member_removed_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "add-member-error",
                      G_CALLBACK (tapiocaui_shell_session_add_member_er_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "state-changed",
                      G_CALLBACK (tapiocaui_shell_session_state_changed_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "error",
                      G_CALLBACK (tapiocaui_shell_session_error_signal_cb),
                      self);

    comm = tapiocaui_shell_get_communication (self, from); 

    if (type == TAPIOCA_SESSION_TYPE_AUDIO) {
        gchar *local_cap = NULL, *remote_cap = NULL;

        comm->audio_session = TAPIOCA_AUDIO_SESSION (session);

        setup_rtp_session (self, comm);
        setup_rtp_stream (self, comm);
    
        farsight_stream_prepare_transports (comm->stream);

        if (tapioca_audio_session_get_capabilities (comm->audio_session,
                    &local_cap,
                    &remote_cap)) {
            remote_capabilities_updated_cb (
                    comm->audio_session, 
                    remote_cap, 
                    comm);
        }
        
        /* Connect audio session signals */
        g_signal_connect (G_OBJECT (comm->audio_session),
                          "remote-capabilities-updated",
                          G_CALLBACK (remote_capabilities_updated_cb),
                          comm);
        g_signal_connect (G_OBJECT (comm->audio_session),
                          "remote-candidate-added",
                          G_CALLBACK (remote_candidate_added_cb),
                          comm);
        
        tapiocaui_communication_set_status (comm->comm_win, TAPIOCAUI_CALL_INCOMING);

        g_free (local_cap);
        g_free (remote_cap);
    }
    else {
        TapiocaTextSession *text_session = TAPIOCA_TEXT_SESSION (session);
        comm->text_sessions = g_list_append (comm->text_sessions, text_session);

        /* Connect text session signals */
        g_signal_connect (G_OBJECT (text_session),
                          "incoming",
                          G_CALLBACK (tapiocaui_shell_text_session_incoming_cb),
                          self);
        g_signal_connect (G_OBJECT (text_session),
                          "delivery-error",
                          G_CALLBACK (tapiocaui_shell_text_session_derror_cb),
                          self);
        tapioca_text_session_disable_message_queue (text_session);
    }
}

/* here will be processed all connection errors */
static void
tapiocaui_shell_connection_error_cb (TapiocaConnection *tc,
                                     guint code,
                                     gpointer user_data)
{
    tapiocaui_debug ("");
}

static void
tapiocaui_shell_session_member_added_cb (TapiocaSession *session,
                                         const gchar *uri,
                                         gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

static void
tapiocaui_shell_session_member_removed_cb (TapiocaSession *session,
                                           const gchar *uri,
                                           gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

static void
tapiocaui_shell_session_add_member_er_cb (TapiocaSession *session,
                                          guint code,
                                          gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

static void
tapiocaui_shell_session_state_changed_cb (TapiocaSession *session,
                                          guint state,
                                          gpointer user_data)
{
    gchar **session_members = NULL;
    TapiocaUIShellActiveComm *active_comm = NULL;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    guint type;

    session_members = tapioca_session_get_members (session);

    if (!session_members) {
        tapiocaui_debug ("No member in session");
        return;
    }

    active_comm = tapiocaui_shell_get_communication (self, session_members[0]);

    type = tapioca_session_get_session_type (session);                          

    switch (state) {
        case TAPIOCA_SESSION_STATE_INITIATING:
            break;
        case TAPIOCA_SESSION_STATE_ESTABLISHED:
            if (type == TAPIOCA_SESSION_TYPE_AUDIO) {
                tapiocaui_shell_open_call (self, active_comm);
            }
            break;
        case TAPIOCA_SESSION_STATE_CLOSED:
            tapiocaui_debug ("received session state closed");
            if (type == TAPIOCA_SESSION_TYPE_AUDIO) {
                tapiocaui_shell_close_call (self, active_comm);
            }
            else if (type == TAPIOCA_SESSION_TYPE_TEXT) {
                tapiocaui_communication_closed_win_message (active_comm->comm_win,
                                                            active_comm->buddy->name);
                active_comm->text_sessions = g_list_remove (active_comm->text_sessions, session);
            }
            g_object_unref (session);
            break;
        
        default:
            break;
    }
    
    g_free (session_members);
}

static void
tapiocaui_shell_session_error_signal_cb (TapiocaSession *session,
                                         guint code,
                                         gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

/* here the media capabilitie will be processed */
static void
remote_capabilities_updated_cb (TapiocaAudioSession *session,
                                const gchar *remote_cap,
                                gpointer user_data)
{
    TapiocaUIShellActiveComm *comm = (TapiocaUIShellActiveComm *) user_data;
    GList *codecs;

    tapiocaui_debug ("remote capabilities updated to %s", remote_cap);

    g_mutex_lock (comm->mutex);
    
    /* TODO check if its needed to stop and start the 
     *      pipeline if it is already running.
     */
    codecs = create_codecs_from_sdp (remote_cap);
    farsight_stream_set_remote_codecs (comm->stream, codecs);
    farsight_codec_list_destroy (codecs);

    comm->remote_codecs_set = TRUE;
    
    if (comm->active_candidate_set) {
        if (farsight_stream_get_state (comm->stream) == FARSIGHT_STREAM_STATE_STOPPED) {
            tapiocaui_debug ("starting media stream ...");
            farsight_stream_start (comm->stream);
        }
    }

    g_mutex_unlock (comm->mutex);
}

static void
tapiocaui_shell_text_session_incoming_cb (TapiocaTextSession *session,
                                          const gchar        *from,
                                          guint               type,
                                          const gchar        *subject,
                                          const gchar        *body,
                                          guint               timestamp,
                                          gpointer            user_data)
{
    TapiocaUIShellActiveComm *comm = NULL;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    tapiocaui_debug ("");
    if (type == TAPIOCA_TEXT_SESSION_MESSAGE_TYPE_PLAIN_TEXT) {
        comm = tapiocaui_shell_get_communication (self, from);
        if (comm) {
            comm->current_text_session = session;    
            tapiocaui_communication_append_message (comm->comm_win, timestamp, comm->buddy->name, body);
        }
    }
    // TODO TAPIOCA_TEXT_SESSION_MESSAGE_TYPE_IS_TYPING
}

static void
tapiocaui_shell_text_session_derror_cb (TapiocaTextSession *session,
                                        const gchar        *from,
                                        guint               type,
                                        const gchar        *subject,
                                        const gchar        *body,
                                        guint               timestamp,
                                        guint               code,
                                        gpointer            user_data)
{
    tapiocaui_debug ("Not implemented");
}

static void
tapiocaui_shell_presence_status_up_cb (TapiocaPresence *presence,
                                       const gchar *jid,
                                       guint status,
                                       const gchar *status_desc,
                                       gboolean voice_cap,
                                       gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    gchar *bare_jid;
    TapiocaUIBuddyData *buddy;
    TapiocaUIBuddyPresence *tmp;
    gchar *aux = NULL;
    
    if (strchr (jid, '/'))
        bare_jid = g_strndup (jid, strchr (jid, '/') - jid);
    else
        bare_jid = g_strdup (jid);

    tapiocaui_debug ("presence status for contact %s - status=%d", jid, status);

    buddy = tapiocaui_db_find_buddy_by_id (self->priv->db_buddy,
                                           TAPIOCAUI_DB_ENTRY_BUDDY,
                                           bare_jid);
    if (buddy) {
        if (status == TAPIOCA_PRESENCE_STATUS_OFFLINE)
            g_hash_table_remove (buddy->jids, jid);
        else {
            tmp = g_hash_table_lookup (buddy->jids, jid);
            if (!tmp) {
                tmp = g_new0 (TapiocaUIBuddyPresence, 1);
                g_hash_table_insert (buddy->jids, g_strdup (jid), tmp);
            }
            if (tmp->status_desc)
                g_free (tmp->status_desc);

            tmp->status = status;

            tmp->voice_cap = voice_cap;

            tapiocaui_shell_set_buddy_presence_priority (tmp);

            if (!status_desc) 
                tmp->status_desc = g_strdup (presence_status_desc [status]); 
            else {
                aux = g_strdup (status_desc);
                aux = g_strstrip (aux);
                if (strlen (aux) == 0) {
                    tmp->status_desc = g_strdup (presence_status_desc [status]);        
                    g_free (aux);
                }
                else 
                    tmp->status_desc = aux;
            }
        }
                
        tapiocaui_main_win_update_buddy (TAPIOCAUI_MAIN_WIN (self->priv->win));
    }
    else {
        tapiocaui_debug ("internal error");
    }

    g_free (bare_jid);
}

static void
tapiocaui_shell_presence_registered_cb (TapiocaPresence *presence,
                                        gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

static void
tapiocaui_shell_presence_unregistered_cb (TapiocaPresence *presence,
                                          gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

static void
tapiocaui_shell_contact_list_contact_authorized_cb (TapiocaContactList *tcl,
                                                    const gchar        *addr,
                                                    gpointer            user_data)
{
    TapiocaUIBuddyData *buddy = NULL;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    gchar *user_id = g_strdup (addr);

    buddy = tapiocaui_db_find_buddy_by_id (self->priv->db_buddy, TAPIOCAUI_DB_ENTRY_BUDDY, user_id);
    
    if (buddy)
        buddy->subscription_state = TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_BOTH;
            
    if (self->priv->presence) {
        gchar *contact[] = { user_id, NULL };
        tapiocaui_debug ("request presence for user %s", user_id);
        if (!(tapioca_presence_request_status (self->priv->presence, (const gchar **) contact))) {
            tapiocaui_debug ("failed on request presence");
        }
    }
    g_free (user_id);
}

static void
tapiocaui_shell_contact_list_contact_denied_cb (TapiocaContactList *tcl,
                                                const gchar        *addr,
                                                gpointer            user_data)
{
    TapiocaUIBuddyData *buddy = NULL;
    gchar *user_id = g_strdup (addr);
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    GtkWidget *dialog = gtk_message_dialog_new_with_markup (GTK_WINDOW (self->priv->win),
                                                            GTK_DIALOG_MODAL,
                                                            GTK_MESSAGE_INFO,
                                                            GTK_BUTTONS_NONE,
                                                            _("The user <b>%s</b> has denied your authorization request"),
                                                            user_id);

    gtk_dialog_add_buttons (GTK_DIALOG (dialog),
                            GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);
    
    gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);

    gtk_dialog_run (GTK_DIALOG (dialog));

    buddy = tapiocaui_db_find_buddy_by_id (self->priv->db_buddy, TAPIOCAUI_DB_ENTRY_BUDDY, user_id);

    gtk_widget_destroy (GTK_WIDGET (dialog));

    if (buddy)
        tapiocaui_db_remove_buddy (self->priv->db_buddy, buddy);

    g_free (user_id);
}

static void
tapiocaui_shell_contact_list_auth_requested_cb (TapiocaContactList  *tcl,
                                                const gchar         *addr,
                                                gpointer             user_data)
{
    guint response;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    gchar *user_id = g_strdup (addr);

    GtkWidget *dialog = gtk_message_dialog_new_with_markup (GTK_WINDOW (self->priv->win),
                                                            GTK_DIALOG_MODAL,
                                                            GTK_MESSAGE_QUESTION,
                                                            GTK_BUTTONS_NONE,
                                                            _("The user <b>%s</b> has added you to his/her contact list. Authorize?"),
                                                            user_id);

    gtk_dialog_add_buttons (GTK_DIALOG (dialog),
                            GTK_STOCK_NO, GTK_RESPONSE_CANCEL,
                            GTK_STOCK_YES, GTK_RESPONSE_OK, NULL);
    
    gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);

    response = gtk_dialog_run (GTK_DIALOG (dialog));

    switch (response) {
        case GTK_RESPONSE_OK:
            if (!tapioca_contact_list_authorize_contact (self->priv->contact_list,
                                                         user_id, ""))
                tapiocaui_debug ("Failed to authorize contact");

                TapiocaUIBuddyData *buddy = tapiocaui_db_entry_new (TAPIOCAUI_DB_ENTRY_BUDDY);
                buddy->id = g_strdup (user_id);
                buddy->name = g_strdup (user_id);
                buddy->group = g_strdup ("");
                buddy->subscription_state = TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_BOTH;
                tapiocaui_db_insert_buddy (self->priv->db_buddy, buddy);
            break;
        default:
            if (!tapioca_contact_list_deny_contact (self->priv->contact_list,
                                                    user_id))
                tapiocaui_debug ("Failed to deny contact");
            break;
    }

    gtk_widget_destroy (GTK_WIDGET (dialog));

    g_free (user_id);
}

static void
tapiocaui_shell_contact_list_contacts_retrieved_cb (TapiocaContactList  *tcl,
                                                    const gchar         *addr,
                                                    const gchar         *alias,
                                                    const gchar         *group,
                                                    guint                state,
                                                    gpointer             user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    TapiocaUIBuddyData *buddy = NULL;

    buddy = tapiocaui_db_entry_new (TAPIOCAUI_DB_ENTRY_BUDDY);
    buddy->id = g_strdup (addr);
    buddy->name = g_strdup (alias);
    buddy->group = g_strdup (group);
    buddy->subscription_state = state;

    tapiocaui_debug ("id=%s name=%s group=%s state=%d", buddy->id, buddy->name, buddy->group, buddy->subscription_state);

    /* Syncronize remote buddy with local buddys */
    tapiocaui_db_sync_buddy (self->priv->db_buddy, buddy);

    /* Request presence for buddy */
    if (self->priv->presence) {
        gchar *contact[] = { buddy->id, NULL };
        tapiocaui_debug ("request presence for user %s", buddy->id);
        if (!(tapioca_presence_request_status (self->priv->presence, (const gchar **) contact))) {
            tapiocaui_debug ("failed on request presence");
        }
    }
    tapiocaui_db_commit (self->priv->db_buddy);
}

static void
tapiocaui_shell_contact_list_contact_info_retrieved_cb (TapiocaContactList  *tcl,
                                                        const gchar         *addr,
                                                        gpointer            *info,
                                                        gpointer             user_data)
{
    tapiocaui_debug ("Not implemented");
}

static void
tapiocaui_shell_buddy_added_cb (TapiocaUIMainWin *win,
                                const gchar *user,
                                gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    gchar *user_id = g_strdup (user);

    if (self->priv->contact_list) {
        TapiocaUIBuddyData *buddy = tapiocaui_db_find_buddy_by_id (self->priv->db_buddy,
                                                                   TAPIOCAUI_DB_ENTRY_BUDDY,
                                                                   user_id);
        if (buddy) {
            if (!tapioca_contact_list_add_contact (self->priv->contact_list,
                                                   buddy->id,
                                                   buddy->name,
                                                   buddy->group))
                tapiocaui_debug ("failed on add contact");

            buddy->subscription_state = TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_NONE;
        }
    }
    if (self->priv->presence) {
        gchar *contact[] = { user_id, NULL };
        tapiocaui_debug ("request presence for user %s", user_id);
        if (!(tapioca_presence_request_status (self->priv->presence, (const gchar **) contact))) {
            tapiocaui_debug ("failed on request presence");
        }
    }
    g_free (user_id);
}

static void
tapiocaui_shell_buddy_removed_cb (TapiocaUIMainWin *win,
                                  const gchar *user,
                                  gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    gchar * user_id = g_strdup (user);

    if (self->priv->contact_list) {
        if (!tapioca_contact_list_remove_contact (self->priv->contact_list, user_id))
            tapiocaui_debug ("Failed on remove contact from remote contact list");
    }

    if (self->priv->presence) {
        gchar *contact[] = { user_id, NULL };
        tapiocaui_debug ("release presence for user", user_id);
        if (!(tapioca_presence_release_status (self->priv->presence, (const gchar **) contact))) {
            tapiocaui_debug ("failed on release presence");
        }
    }
    g_free (user_id);
}

static void
tapiocaui_shell_buddy_updated_cb (TapiocaUIMainWin *win,
                                  const gchar *old_user,
                                  const gchar *new_user,
                                  gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    if (self->priv->presence) {
        gchar *contact[] = { g_strdup (old_user), NULL };
        tapiocaui_debug ("release presence for user %s", old_user);
        if (!(tapioca_presence_release_status (self->priv->presence, (const gchar **) contact))) {
            tapiocaui_debug ("failed on release presence");
        }
        
        g_free (contact[0]);
        contact[0] = g_strdup (new_user);

        tapiocaui_debug ("request presence for user %s", new_user);
        if (!(tapioca_presence_request_status (self->priv->presence, (const gchar **) contact))) {
            tapiocaui_debug ("failed on request presence");
        }
        g_free (contact[0]);
    }
}

static void
tapiocaui_shell_account_status_updated_cb (TapiocaUIMainWin *win,
                                           guint status,
                                           const gchar *status_desc,
                                           gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    if (self->priv->presence) {
        if (!(tapioca_presence_set_status (self->priv->presence, status, status_desc))) {
            tapiocaui_debug ("failed setting presence status");
            return;
        }
    }
}

static gboolean
tapiocaui_shell_call_buddy_find (gpointer key,
                                 gpointer value,
                                 gpointer data)
{
    TapiocaAudioSession *audio = TAPIOCA_AUDIO_SESSION (data);
    gchar **members = g_new0 (gchar *, 2);

    members[0] = g_strdup (key);

    if (!(tapioca_session_add_members (TAPIOCA_SESSION (audio), 
                                       (const gchar **) members))) {
        tapiocaui_debug ("error on add_member %s", members[0]);
        g_strfreev (members);
        return FALSE;
     }

    if (!tapioca_session_initiate (TAPIOCA_SESSION (audio))) {
        tapiocaui_debug ("error on initiate audio session with member %s", members[0]);
        tapioca_session_remove_members (TAPIOCA_SESSION (audio),
                                        (const gchar **) members);
        g_strfreev (members);
        return FALSE;
    }

    g_strfreev (members);

    return TRUE;
}

/* NAT traversal stuff */
static void
setup_rtp_session (TapiocaUIShell *self,
                   TapiocaUIShellActiveComm *comm)
{
    comm->session = farsight_session_factory_make ("rtp");
    if (!comm->session) {
        g_error ("RTP plugin not found");
        exit (1);
    }
    
    tapiocaui_debug ("protocol details:\n name: %s\n description: %s\n author: %s\n",
            farsight_plugin_get_name (comm->session->plugin),
            farsight_plugin_get_description (comm->session->plugin),
            farsight_plugin_get_author (comm->session->plugin));
    g_signal_connect (G_OBJECT (comm->session), "error", 
            G_CALLBACK (session_error), NULL);
}

static void
setup_rtp_stream (TapiocaUIShell *self, 
                  TapiocaUIShellActiveComm *comm)
{
    const GList *possible_codecs, *lp;
    FarsightCodec *codec;
    GstElement *srcbin;
    GstElement *alsasrc;
    GstPad *srcpad;
    /* GstElement *sinkbin; */
    GstElement *alsasink;
    /* GstPad *sinkpad; */

    comm->stream = farsight_session_create_stream (comm->session,
            FARSIGHT_MEDIA_TYPE_AUDIO, FARSIGHT_STREAM_DIRECTION_BOTH);
    tapiocaui_debug ("setting up stream %p with self %p and comm %p",
            comm->stream, self, comm);
    g_signal_connect (G_OBJECT (comm->stream), "error", 
                      G_CALLBACK (stream_error), self);
    g_signal_connect (G_OBJECT (comm->stream), "new-active-candidate-pair", 
                      G_CALLBACK (new_active_candidate_pair), comm);
    g_signal_connect (G_OBJECT (comm->stream), "codec-changed", 
                      G_CALLBACK (codec_changed), comm);
    g_signal_connect (G_OBJECT (comm->stream), "native-candidates-prepared", 
                      G_CALLBACK (native_candidates_prepared), comm);
    g_signal_connect (G_OBJECT (comm->stream), "new-native-candidate",
                      G_CALLBACK (new_native_candidate), comm);
    g_signal_connect (G_OBJECT (comm->stream), "state-changed", 
                      G_CALLBACK (state_changed), self);

    possible_codecs = farsight_stream_get_local_codecs (comm->stream);
    for (lp = possible_codecs; lp; lp = g_list_next (lp)) {
        codec = (FarsightCodec*) lp->data;
        tapiocaui_debug ("codec: %d: %s/%d found", 
                codec->id, codec->encoding_name, 
                codec->clock_rate);
    }

    /* create src bin */
    srcbin = gst_bin_new ("src");
    alsasrc = gst_element_factory_make ("alsasrc", "alsasrc"); 
    g_object_set (G_OBJECT (alsasrc), "blocksize", 320, NULL);
    g_object_set (G_OBJECT (alsasrc), "latency-time", G_GINT64_CONSTANT (20000), NULL);
    /* default output volume to 100% */
    comm->srcvolume = gst_element_factory_make ("volume", "srcvolume");
    g_object_set (G_OBJECT (comm->srcvolume), "volume", 4.0, NULL);
    gst_bin_add_many (GST_BIN (srcbin), alsasrc, comm->srcvolume, NULL);
    gst_element_link (alsasrc, comm->srcvolume);
    srcpad = gst_ghost_pad_new ("src", 
            gst_element_get_pad (comm->srcvolume, "src"));
    gst_element_add_pad (srcbin, srcpad);
  
    /* create sink bin */
    /* sinkbin = gst_bin_new ("sink"); */
    /* default output volume to 50% */
    /*
    comm->sinkvolume = gst_element_factory_make ("volume", "sinkvolume");
    g_object_set (G_OBJECT (comm->sinkvolume), "volume", 2.0, NULL);
    */
    alsasink = gst_element_factory_make ("alsasink", "alsasink");
    g_object_set (G_OBJECT (alsasink), "sync", FALSE, NULL);
    g_object_set (G_OBJECT (alsasink), "latency-time", G_GINT64_CONSTANT (20000), NULL);
    g_object_set (G_OBJECT (alsasink), "buffer-time", G_GINT64_CONSTANT (80000), NULL);
    /*
    gst_bin_add_many (GST_BIN (sinkbin), comm->sinkvolume, alsasink, NULL);
    gst_element_link (comm->sinkvolume, alsasink);
    sinkpad = gst_ghost_pad_new ("sink", 
            gst_element_get_pad (comm->sinkvolume, "sink"));
    gst_element_add_pad (sinkbin, sinkpad); 
    */

    farsight_stream_set_source (comm->stream, srcbin);
    /* farsight_stream_set_sink (comm->stream, sinkbin); */
    farsight_stream_set_sink (comm->stream, alsasink);
}

static void
close_rtp_session (TapiocaUIShell *self,
                   TapiocaUIShellActiveComm *comm)
{
    FarsightSession *session = comm->session;

    tapiocaui_debug ("stoping media session!");
    g_signal_handlers_disconnect_by_func (G_OBJECT (session), G_CALLBACK (session_error), NULL);
    g_object_unref (session);
    comm->session = NULL;
}

static void
close_rtp_stream (TapiocaUIShell *self,
                  TapiocaUIShellActiveComm *comm)
{
    FarsightStream *stream = comm->stream;
    
    tapiocaui_debug ("stoping media stream!");
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (stream_error), self);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (new_active_candidate_pair), comm);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (codec_changed), comm);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (native_candidates_prepared), comm);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (new_native_candidate), comm);
    g_signal_handlers_disconnect_by_func (G_OBJECT (stream), G_CALLBACK (state_changed), self);

    if (farsight_stream_get_state != FARSIGHT_STREAM_STATE_STOPPED)
        farsight_stream_stop (stream);
    g_object_unref (stream);
    comm->stream = NULL;
}

static void
remote_candidate_added_cb (TapiocaAudioSession *audio_session, 
                           const gchar *name,
                           const gchar *protocol,
                           const gchar *ip,
                           guint port,
                           gdouble preference,
                           const gchar *username,
                           const gchar *password,
                           const gchar *type,
                           guint generation,
                           gpointer user_data)
{
    TapiocaUIShellActiveComm *comm = (TapiocaUIShellActiveComm *) user_data;
    GList *candidate = NULL;
    FarsightTransportInfo *info;

    info = g_new0 (FarsightTransportInfo, 1);
    info->candidate_id = g_strdup_printf ("L%d", ++comm->candidate_id);
    info->component = 1;
    info->ip = g_strdup (ip);
    info->port = port;
    info->proto = (g_strcasecmp (protocol, "tcp") == 0) ? FARSIGHT_NETWORK_PROTOCOL_TCP : FARSIGHT_NETWORK_PROTOCOL_UDP;
    info->proto_subtype = g_strdup ("RTP");
    info->proto_profile = g_strdup ("AVP");
    info->preference = preference;
    if (!g_strcasecmp (type, "local"))
        info->type = FARSIGHT_CANDIDATE_TYPE_LOCAL;
    else if (!g_strcasecmp (type, "stun"))
        info->type = FARSIGHT_CANDIDATE_TYPE_DERIVED;
    else if (!g_strcasecmp (type, "relay"))
        info->type = FARSIGHT_CANDIDATE_TYPE_RELAY;
    info->username = g_strdup (username);
    info->password = g_strdup (password);
 
    tapiocaui_debug ("remote transport candidate: %s %d %s (%d) %s %s:%d %s %s, pref %f",
                info->candidate_id, 
                info->component,
                protocol,
                info->proto,
                info->proto_subtype, 
                info->ip,
                info->port,
                info->username, 
                info->password,
                (double) info->preference);
                      
    candidate = g_list_append (candidate, info);
    farsight_stream_add_remote_candidate (comm->stream, candidate);

    farsight_transport_list_destroy (candidate);
}

static void
session_error (FarsightSession *stream, 
               FarsightSessionError error, 
               const gchar *debug, 
               gpointer user_data)
{
    tapiocaui_debug ("session error: session=%p error=%s", stream, debug);
}

static void
stream_error (FarsightStream *stream, 
              FarsightStreamError error, 
              const gchar *debug, 
              gpointer user_data)
{
    tapiocaui_debug ("stream error: stream=%p error=%s", stream, debug);
}

static void
new_active_candidate_pair (FarsightStream *stream, 
                           gchar *native_candidate, 
                           gchar *remote_candidate, 
                           gpointer user_data)
{
    TapiocaUIShellActiveComm *comm = (TapiocaUIShellActiveComm *) user_data;

    tapiocaui_debug ("new native candidate pair: stream=%p", stream);

    g_mutex_lock (comm->mutex);
    
    if (comm->remote_codecs_set) {
        if (farsight_stream_get_state (stream) == FARSIGHT_STREAM_STATE_STOPPED) {
            tapiocaui_debug ("starting media stream ...");
            farsight_stream_start (stream);
        }
    }
    else {
        comm->active_candidate_set = TRUE;
    }
    
    g_mutex_unlock (comm->mutex);
}

static void
codec_changed (FarsightStream *stream, 
               gint codec_id, 
               gpointer user_data)
{
    tapiocaui_debug ("codec changed: codec_id=%d, stream=%p", codec_id, stream);
}

static void
new_native_candidate (FarsightStream *stream, 
                      const gchar *candidate_id, 
                      gpointer user_data)
{
    TapiocaUIShellActiveComm *comm = (TapiocaUIShellActiveComm *) user_data;
    GList *native_candidate = NULL, *lc;
    FarsightTransportInfo *info;
    gchar *proto, *type = NULL;

    tapiocaui_debug ("new native candidate found");

    native_candidate = farsight_stream_get_native_candidate (stream, candidate_id);

    for (lc = native_candidate; lc; lc = g_list_next (lc)) {
        info = (FarsightTransportInfo *) lc->data;
        proto = (info->proto == FARSIGHT_NETWORK_PROTOCOL_TCP) ? "tcp" : "udp";
        if (info->type == FARSIGHT_CANDIDATE_TYPE_LOCAL)
            type = "local";
        else if (info->type == FARSIGHT_CANDIDATE_TYPE_DERIVED)
            type = "stun";
        else if (info->type == FARSIGHT_CANDIDATE_TYPE_RELAY)
            type = "relay";
        
        tapiocaui_debug ("local transport candidate: %s %d %s %s %s:%d %s %s, pref %f",
                info->candidate_id, 
                info->component,
                proto,
                info->proto_subtype, info->ip,
                info->port,
                info->username, info->password,
                (double) info->preference);
        tapioca_audio_session_add_local_candidate (
                comm->audio_session,
                "rtp",
                // info->candidate_id,
                proto,
                info->ip,
                info->port,
                info->preference,
                info->username,
                info->password,
                type,
                0);
    }

    farsight_transport_list_destroy (native_candidate);
}

static void
native_candidates_prepared (FarsightStream *stream, 
                            gpointer user_data)
{
    const GList *transport_candidates, *lp;
    FarsightTransportInfo *info;

    tapiocaui_debug ("candidates preparation complete: stream=%p", stream);

    transport_candidates = farsight_stream_get_native_candidate_list (stream); 
    for (lp = transport_candidates; lp; lp = g_list_next (lp)) 
    {
        info = (FarsightTransportInfo*)lp->data;
        tapiocaui_debug ("local transport candidate: %s %d %s %s %s:%d, pref %f", 
                info->candidate_id, info->component, 
                (info->proto == FARSIGHT_NETWORK_PROTOCOL_TCP) ? "TCP" : "UDP",
                info->proto_subtype, info->ip, info->port, 
                (double) info->preference);
    }
}

static void 
state_changed (FarsightStream *stream, 
               FarsightStreamState state, 
               FarsightStreamDirection dir, 
               gpointer user_data)
{
    TapiocaUIShell *self = (TapiocaUIShell *) user_data;

    switch (state) {
        case FARSIGHT_STREAM_STATE_STOPPED:
            tapiocaui_debug ("stream stopped: stream=%p", stream);
            tapiocaui_shell_close_call (self, self->priv->active_comm);
            break;
        case FARSIGHT_STREAM_STATE_PLAYING: 
            tapiocaui_debug ("stream playing: stream=%p", stream);
            break;
    }
}

static GList *
create_codecs_from_sdp (const gchar *sdp_)
{
    GList *codecs = NULL;
    FarsightCodec *codec;
    gchar *sdp;
    guint pt;
    gint rate, channels;
    gchar *encoding_name, *offset;
   
    // TODO check if the string is correctly formed

    // a=rtpmap:PT ENCODING_NAME/RATE[/CHANNELS]\r\n";
    sdp = strstr (sdp_, "a=rtpmap:");
    while (sdp) {
        sdp = strchr (sdp, ':') + 1;    // sdp = "pt ENCODING_NAME/RATE[/CHANNELS]
        pt = strtoul (sdp, &sdp, 10);   // sdp = " ENCODING_NAME/RATE[/CHANNELS]
        ++sdp;                          // sdp = "ENCODING_NAME/RATE[/CHANNELS]
        offset = strchr (sdp, '/');
        encoding_name = g_strndup (sdp, offset - sdp);
        sdp = offset + 1;              // sdp = "RATE[/CHANNELS]
        rate = strtoul (sdp, &sdp, 10); // sdp = " [/CHANNELS]
        if (*sdp == '/') {
            ++sdp;                      // sdp = CHANNELS
            channels = strtoul (sdp, NULL, 10);
        }
        else
            channels = -1;

        if (channels != -1)
            g_debug ("adding codec %d %s/%d/%d", pt, encoding_name, rate, channels);
        else
            g_debug ("adding codec %d %s/%d", pt, encoding_name, rate);

        codec = g_new0 (FarsightCodec, 1);
        farsight_codec_init (codec, pt, encoding_name, FARSIGHT_MEDIA_TYPE_AUDIO, rate);
        codecs = g_list_append (codecs, codec);

        g_free (encoding_name);

        sdp = strstr (sdp, "a=rtpmap:");
    }

    return codecs;
}

static gchar *
create_sdp_from_codecs (const GList *codecs)
{
    const GList *lc;
    FarsightCodec *codec;
    GString *sdp = g_string_new ("v=0\r\n"
            "o=- 123456 654321 IN IP4 127.0.0.1\r\n"
            "c=IN IP4 127.0.0.1\r\n");

    for (lc = codecs; lc; lc = g_list_next (lc)) {
        codec = (FarsightCodec *) lc->data;
        if (codec->media_type != FARSIGHT_MEDIA_TYPE_AUDIO)
            continue;
        
        g_string_append_printf (sdp, "a=rtpmap:%d %s/%d\r\n", 
                codec->id, codec->encoding_name, codec->clock_rate);
    }
    
    return g_string_free (sdp, FALSE);
}

static void
tapiocaui_shell_set_buddy_presence_priority (TapiocaUIBuddyPresence *buddy_presence)
{
    g_return_if_fail (buddy_presence != NULL);

    switch (buddy_presence->status) {
        case TAPIOCA_PRESENCE_STATUS_AVAILABLE:
            buddy_presence->priority = 0;
            break;
        case TAPIOCA_PRESENCE_STATUS_DND:
            buddy_presence->priority = 10;
            break;
        case TAPIOCA_PRESENCE_STATUS_AWAY:
            buddy_presence->priority = 20;
            break;
        case TAPIOCA_PRESENCE_STATUS_XA:
            buddy_presence->priority = 30;
            break;
        case TAPIOCA_PRESENCE_STATUS_OFFLINE:
            buddy_presence->priority = 40;
            break;
        default:
            buddy_presence->priority = 50;
            break;
    }
}
